(function (global, factory) {
typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory(require('chart.js'), require('chart.js/helpers')) :
typeof define === 'function' && define.amd ? define(['chart.js', 'chart.js/helpers'], factory) :
(global = typeof globalThis !== 'undefined' ? globalThis : global || self, global["chartjs-plugin-outlabels3"] = factory(global.Chart, global.Chart.helpers));
})(this, (function (chart_js, helpers) { 'use strict';

class PaddingOptions {
    constructor() {
        this.top = 4;
        this.right = 4;
        this.bottom = 4;
        this.left = 4;
    }
}
class FontOptions {
    constructor() {
        this.family = "'Helvetica Neue', 'Helvetica', 'Arial', sans-serif";
        this.style = 'normal';
        this.resizable = true;
        this.lineHeight = 1.2;
        this.lineSize = 0;
    }
}
class OutLabelsOptions {
    constructor() {
        this.display = true;
        this.text = '%l %p';
        this.textAlign = 'center';
        this.color = 'white';
        this.borderRadius = 0;
        this.borderWidth = 0;
        this.lineWidth = 2;
        this.length = 40;
        this.percentPrecision = 1;
        this.valuePrecision = 3;
        this.padding = new PaddingOptions();
        this.font = new FontOptions();
    }
    backgroundColor(context) {
        return context.dataset.backgroundColor;
    }
    borderColor(context) {
        return context.dataset.backgroundColor;
    }
    lineColor(context) {
        return context.dataset.backgroundColor;
    }
}

function getFontString(font) {
    if (!font || helpers.isNullOrUndef(font.size) || helpers.isNullOrUndef(font.family)) {
        return '';
    }
    return ((font.style ? font.style + ' ' : '') +
        (font.weight ? font.weight + ' ' : '') +
        font.size +
        'px ' +
        font.family);
}
// @todo move this in Chart.helpers.canvas.textSize
// @todo cache calls of measureText if font doesn't change?!
function textSize(ctx, lines, font) {
    const prev = ctx.font;
    let width = 0;
    ctx.font = getFontString(font);
    for (let i = 0; i < lines.length; ++i) {
        width = Math.max(ctx.measureText(lines[i]).width, width);
    }
    ctx.font = prev;
    return {
        height: lines.length * font.lineSize,
        width: width,
    };
}
// @todo move this method in Chart.helpers.options.toFont
function parseFont(value, height) {
    const defaults = chart_js.Chart.defaults;
    let size = helpers.valueOrDefault(value.size, defaults.font.size);
    if (value.resizable) {
        size = adaptTextSizeToHeight(height, value.minSize, value.maxSize);
    }
    value.lineSize = helpers.toLineHeight(value.lineHeight.toString(), size !== null && size !== void 0 ? size : 0);
    value.size = size;
    return value;
}
function adaptTextSizeToHeight(height, min, max) {
    const size = (height / 100) * 2.5;
    if (min && size < min) {
        return min;
    }
    if (max && size > max) {
        return max;
    }
    return size;
}
function drawRoundedRect(ctx, x, y, w, h, radius) {
    const HALF_PI = Math.PI / 2;
    if (radius) {
        const r = Math.min(radius, h / 2, w / 2);
        const left = x + r;
        const top = y + r;
        const right = x + w - r;
        const bottom = y + h - r;
        ctx.moveTo(x, top);
        if (left < right && top < bottom) {
            ctx.arc(left, top, r, -Math.PI, -HALF_PI);
            ctx.arc(right, top, r, -HALF_PI, 0);
            ctx.arc(right, bottom, r, 0, HALF_PI);
            ctx.arc(left, bottom, r, HALF_PI, Math.PI);
        }
        else if (left < right) {
            ctx.moveTo(left, y);
            ctx.arc(right, top, r, -HALF_PI, HALF_PI);
            ctx.arc(left, top, r, HALF_PI, Math.PI + HALF_PI);
        }
        else if (top < bottom) {
            ctx.arc(left, top, r, -Math.PI, 0);
            ctx.arc(left, bottom, r, 0, Math.PI);
        }
        else {
            ctx.arc(left, top, r, -Math.PI, Math.PI);
        }
        ctx.closePath();
        ctx.moveTo(x, y);
    }
    else {
        ctx.rect(x, y, w, h);
    }
}

class OutLabel {
    constructor(ctx, index, config, context) {
        var _a, _b;
        this.rect = { height: 0, width: 0, x: 0, y: 0 };
        if (!config.display) {
            throw new Error("Label display property is set to false.");
        }
        this.ctx = ctx;
        this.index = index;
        this.context = context;
        // Init text
        const label = context.labels[index];
        let text = helpers.resolve([config.text], context, index);
        if (!text)
            text = "%l %p";
        /* Replace label marker */
        text = text.replace(/%l/gi, label);
        /* Replace value marker with possible precision value */
        (text.match(/%v\.?(\d*)/gi) || [])
            .map(function (val) {
            const prec = val.replace(/%v\./gi, "");
            if (prec.length) {
                return +prec;
            }
            else {
                return config.valuePrecision;
            }
        })
            .forEach(function (val) {
            if (text)
                text = text.replace(/%v\.?(\d*)/i, context.value.toFixed(val));
        });
        /* Replace percent marker with possible precision value */
        (text.match(/%p\.?(\d*)/gi) || [])
            .map(function (val) {
            const prec = val.replace(/%p\./gi, "");
            if (prec.length) {
                return +prec;
            }
            else {
                return config.percentPrecision;
            }
        })
            .forEach(function (val) {
            let percentPrecision = config.percentPrecision;
            if (!isNaN(val)) {
                percentPrecision = val;
            }
            if (text)
                text = text.replace(/%p\.?(\d*)/i, (context.percent * 100).toFixed(percentPrecision) + "%");
        });
        // Count lines
        const lines = text.match(/[^\r\n]+/g);
        // If no lines => nothing to display
        if (!lines || !lines.length)
            throw new Error("No text to show.");
        // Remove unnecessary spaces
        for (let i = 0; i < lines.length; ++i) {
            lines[i] = lines[i].trim();
        }
        this.encodedText = config.text;
        this.text = text;
        this.lines = lines;
        this.label = label;
        this.value = context.value;
        // Init style
        const fontConfig = Object.assign(new FontOptions(), config.font);
        this.style = {
            backgroundColor: helpers.resolve([config.backgroundColor, "black"], context, index),
            borderColor: helpers.resolve([config.borderColor, "black"], context, index),
            borderRadius: helpers.resolve([config.borderRadius, 0], context, index),
            borderWidth: helpers.resolve([config.borderWidth, 0], context, index),
            lineWidth: helpers.resolve([config.lineWidth, 2], context, index),
            lineColor: helpers.resolve([config.lineColor, "black"], context, index),
            color: helpers.resolve([config.color, "white"], context, index),
            font: parseFont((_a = helpers.resolve([fontConfig])) !== null && _a !== void 0 ? _a : fontConfig, parseFloat(ctx.canvas.style.height.slice(0, -2))),
            padding: helpers.toPadding(helpers.resolve([config.padding], context, index)),
            textAlign: helpers.resolve([config.textAlign, "left"], context, index),
        };
        this.length = (_b = helpers.resolve([config.length, 40], context, index)) !== null && _b !== void 0 ? _b : 40;
        this.size = textSize(ctx, this.lines, this.style.font);
    }
    get center() {
        return {
            x: this.x,
            y: this.y,
        };
    }
    computeRect() {
        return {
            x: this.x +
                (this.nx < 0 ? -this.size.width : 0) -
                this.style.padding.left -
                this.style.borderWidth,
            y: this.y -
                this.size.height / 2 -
                this.style.padding.top -
                this.style.borderWidth,
            width: this.size.width +
                2 * this.style.borderWidth +
                this.style.padding.width,
            height: this.size.height +
                2 * this.style.borderWidth +
                this.style.padding.height,
        };
    }
    drawText() {
        const font = this.style.font;
        const color = this.style.color;
        if (!this.lines.length || !color) {
            return;
        }
        this.ctx.font = getFontString(this.style.font);
        this.ctx.fillStyle = color;
        this.ctx.textAlign = this.nx < 0 ? "right" : "left";
        this.ctx.textBaseline = "middle";
        const x = this.x;
        let y = this.y;
        for (let idx = 0; idx < this.lines.length; ++idx) {
            this.ctx.fillText(this.lines[idx], Math.round(x), Math.round(y), Math.round(this.size.width));
            y += font.lineSize;
        }
    }
    drawRect() {
        this.ctx.beginPath();
        drawRoundedRect(this.ctx, this.rect.x, this.rect.y, this.rect.width, this.rect.height, this.style.borderRadius);
        this.ctx.closePath();
        if (this.style.backgroundColor) {
            this.ctx.fillStyle = this.style.backgroundColor || "black";
            this.ctx.fill();
        }
        if (this.style.borderColor && this.style.borderWidth) {
            this.ctx.strokeStyle = this.style.borderColor;
            this.ctx.lineWidth = this.style.borderWidth;
            this.ctx.lineJoin = "miter";
            this.ctx.stroke();
        }
    }
    drawLine() {
        this.ctx.save();
        this.ctx.strokeStyle = this.style.lineColor;
        this.ctx.lineWidth = this.style.lineWidth;
        this.ctx.lineJoin = "miter";
        this.ctx.beginPath();
        this.ctx.moveTo(this.x1, this.y1);
        this.ctx.lineTo(this.x, this.y);
        this.ctx.stroke();
        this.ctx.restore();
    }
    draw() {
        this.drawRect();
        this.drawText();
    }
    updateRects() {
        this.rect = this.computeRect();
    }
    positionCenter(arc) {
        this.arc = arc;
        const chart = this.context.chart;
        const midAngle = (arc.startAngle + arc.endAngle) / 2;
        const nx = Math.cos(midAngle);
        const ny = Math.sin(midAngle);
        //const d = arc.outerRadius
        /*
        const viewWidth = chart.chartArea.width
        const viewLeft = chart.chartArea.x
        const viewTop = chart.chartArea.y
        const viewHeight = chart.chartArea.height */
        const cx = (chart.chartArea.left + chart.chartArea.right) / 2;
        const cy = (chart.chartArea.top + chart.chartArea.bottom) / 2;
        const r = arc.outerRadius;
        const x1 = r * nx + cx;
        const y1 = r * ny + cy;
        const x2 = x1 + nx * this.length;
        const y2 = y1 + ny * this.length;
        this.x1 = x1;
        this.y1 = y1;
        this.x = x2;
        this.y = y2;
        this.nx = nx;
    }
}

class OutLabelsManager {
    constructor() {
        this.labels = new Map();
    }
    set(id) {
        this.labels.set(id, new Map());
    }
    get(id) {
        return this.labels.get(id);
    }
    setLabel(id, number, label) {
        const labels = this.get(id);
        if (!labels)
            return;
        labels === null || labels === void 0 ? void 0 : labels.set(number, label);
    }
    adjustQuadrant(list) {
        if (list.length < 2) {
            return false;
        }
        list.sort((i1, i2) => i1.rect.y - i2.rect.y);
        let lastPos = 0;
        let adjusted = false;
        let totalShifts = 0;
        for (let i = 0; i < list.length; i++) {
            const item = list[i];
            const rect = item.rect;
            // eslint-disable-next-line prefer-const
            let delta = rect.y - lastPos;
            if (delta < 0) {
                rect.y -= delta;
                item.y -= delta;
                adjusted = true;
            }
            const shift = Math.max(-delta, 0);
            totalShifts += shift;
            lastPos = rect.y + item.rect.height;
        }
        if (totalShifts > 0) {
            // Shift back to make the distribution more equally.
            const delta = -totalShifts / list.length;
            if (delta !== 0)
                adjusted = true;
            for (let i = 0; i < list.length; i++) {
                const item = list[i];
                const rect = item.rect;
                rect.y += delta;
                item.rect.y += delta;
            }
        }
        return adjusted;
    }
    recalculateX(chart, list) {
        if (list.length < 1)
            return;
        const cx = (chart.chartArea.left + chart.chartArea.right) / 2;
        const cy = (chart.chartArea.top + chart.chartArea.bottom) / 2;
        const r = list[0].arc.outerRadius;
        const dir = list[0].nx < 0 ? -1 : 1;
        let maxY = 0;
        let rB = 0;
        list.forEach(item => {
            const dy = Math.abs(item.rect.y - cy);
            if (dy > maxY) {
                const dx = item.rect.x - cx;
                const rA = r + item.length;
                rB =
                    Math.abs(dx) < rA
                        ? Math.sqrt((dy * dy) / (1 - (dx * dx) / rA / rA))
                        : rA;
                maxY = dy;
            }
        });
        const rB2 = rB * rB;
        list.forEach(item => {
            const dy = Math.abs(item.rect.y - cy);
            // horizontal r is always same with original r because x is not changed.
            const rA = r + item.length;
            const rA2 = rA * rA;
            // Use ellipse implicit function to calculate x
            const dx = Math.sqrt((1 - Math.abs((dy * dy) / rB2)) * rA2);
            const newX = cx + dx * dir;
            item.x = newX;
        });
    }
    avoidOverlap(chart) {
        const labels = this.get(chart.id);
        if (labels) {
            const cx = (chart.chartArea.left + chart.chartArea.right) / 2;
            const cy = (chart.chartArea.top + chart.chartArea.bottom) / 2;
            const topLeftList = [];
            const topRightList = [];
            const bottomLeftList = [];
            const bottomRightList = [];
            labels.forEach(label => {
                if (label.x < cx) {
                    if (label.y < cy)
                        topLeftList.push(label);
                    else
                        bottomLeftList.push(label);
                }
                else {
                    if (label.y < cy)
                        topRightList.push(label);
                    else
                        bottomRightList.push(label);
                }
            });
            if (this.adjustQuadrant(topLeftList))
                this.recalculateX(chart, topLeftList);
            if (this.adjustQuadrant(topRightList))
                this.recalculateX(chart, topRightList);
            if (this.adjustQuadrant(bottomLeftList))
                this.recalculateX(chart, bottomLeftList);
            if (this.adjustQuadrant(bottomRightList))
                this.recalculateX(chart, bottomRightList);
        }
    }
}

const outLabelsManager = new OutLabelsManager();
var OutLabels = {
    id: 'outlabels',
    beforeInit: function (chart) {
        outLabelsManager.set(chart.id);
    },
    afterDatasetUpdate: function (chart, args, options) {
        const config = Object.assign(new OutLabelsOptions(), options);
        const labels = chart.config.data.labels;
        const dataset = chart.data.datasets[args.index];
        const elements = args.meta.data;
        const ctx = chart.ctx;
        ctx.save();
        for (let i = 0; i < elements.length; ++i) {
            const el = elements[i];
            let newLabel = null;
            const percent = dataset.data[i] /
                dataset.data.reduce((sum, current) => sum + current);
            const context = {
                chart: chart,
                dataIndex: i,
                dataset: dataset,
                labels: labels,
                datasetIndex: args.index,
                value: dataset.data[i],
                percent: percent,
            };
            const display = helpers.resolve([config.display, false], context, i);
            if (display && el && chart.getDataVisibility(args.index)) {
                try {
                    newLabel = new OutLabel(ctx, i, config, context);
                }
                catch (e) {
                    console.warn(e);
                    newLabel = null;
                }
            }
            if (newLabel)
                outLabelsManager.setLabel(chart.id, i, newLabel);
        }
        ctx.restore();
    },
    afterDatasetDraw: function (chart, args) {
        const ctx = chart.ctx;
        const elements = args.meta.data;
        ctx.save();
        const chartOutlabels = outLabelsManager.get(chart.id);
        if (!chartOutlabels)
            return;
        chartOutlabels.forEach(label => {
            label.positionCenter(elements[label.index]);
            label.updateRects();
        });
        outLabelsManager.avoidOverlap(chart);
        chartOutlabels.forEach(label => {
            label.updateRects();
            label.draw();
            label.drawLine();
        });
        ctx.restore();
    },
};

chart_js.Chart.register(OutLabels);

return OutLabels;

}));
